﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using IsbnCheckSum.Calculation;
using IsbnCheckSum.Resources;
using IsbnCheckSum.Validation;

namespace IsbnCheckSum
{
	public class IsbnCode : IAmValidatable
	{
		private readonly ICalculateCheckSums calculator;
		private readonly ICanValidateIsbnCodeFormats formatValidator;

		/// <summary>
		/// Instantiates a new <see cref="IsbnCode"/> specifying the concrete strategies
		/// for check sum calculation and format validation.
		/// </summary>
		protected IsbnCode(ICalculateCheckSums calculator, ICanValidateIsbnCodeFormats formatValidator)
		{
			this.calculator = calculator;
			this.formatValidator = formatValidator;
		}

		public IsbnCode(string isbnCode, ICalculateCheckSums calculator, ICanValidateIsbnCodeFormats formatValidator)
			: this(calculator, formatValidator)
		{
			if (this.formatValidator.CodeContainsCheckSum(isbnCode))
			{
				this.BaseSequence = isbnCode.Substring(0, isbnCode.Length - 1);
				this.CheckSumDigit = isbnCode.Last();
			}
			else
			{
				this.BaseSequence = isbnCode;
				this.CalculateCheckSum();
			}
		}

		/// <summary>
		/// Gets the base sequence used to calculate the check sum
		/// </summary>
		public string BaseSequence { get; private set; }

		/// <summary>
		/// Gets the check sum digit
		/// </summary>
		public char CheckSumDigit { get; private set; }

		/// <summary>
		/// Gets the complete ISBN code
		/// </summary>
		public string Code
		{
			get
			{
				return string.Format("{0}{1}", this.BaseSequence, this.CheckSumDigit);
			}
		}

		internal void CalculateCheckSum()
		{
			// Don't override the supplied check digit if there is one
			if (this.CheckSumDigit == Char.MinValue)
			{
				this.CheckSumDigit = this.calculator.CalculateCheckSum(this.BaseSequence);
			}
		}

		#region Validation logic - potential refactor point to avoid violating single responsibility
		public bool IsValid()
		{
			return !this.GetValidationErrors().Any();
		}

		public IEnumerable<string> GetValidationErrors()
		{
			// Return any format errors
			foreach (var error in this.formatValidator.GetValidationErrors(this.Code))
			{
				yield return error;
			}

			// Verify that the check digit matches what it should be
			if(!this.IsCheckDigitCorrect()) 
			{
				yield return IsbnCodeMessages.SuppliedCheckDigitIsNotCorrect;
			}
		}

		private bool IsCheckDigitCorrect()
		{
			// Assume true until proved otherwise.
			bool digitCorrect = true;

			// If the digit isn't set, we don't want to error the validation.
			if (this.CheckSumDigit != Char.MinValue)
			{
				char digitShouldBe = this.calculator.CalculateCheckSum(this.BaseSequence);

				digitCorrect = this.CheckSumDigit == digitShouldBe;
			}

			return digitCorrect;
		}
		#endregion		
	}
}
