﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using IsbnCheckSum.Resources;

namespace IsbnCheckSum.Validation
{
	public class IsbnCodeFormatValidator : ICanValidateIsbnCodeFormats
	{
		private readonly List<int> validLengths;
		private readonly List<int> lengthsRequiringHyphens;
		private readonly List<int> validHyphenCounts;

		public IsbnCodeFormatValidator()
		{
			// Ideally these would be set via config file so that behaviour could be changed at runtime
			// without re-deploying code
			this.validLengths = new List<int> { 10, 13 };
			this.lengthsRequiringHyphens = new List<int> { 13 };
			this.validHyphenCounts = new List<int> { 0, 3 };
		}

		public bool CodeContainsCheckSum(string isbnCode)
		{
			// A bit simplistic, but if the length of supplied code is exactly equal
			// to one of the valid lengths, then assume check sum is included
			return this.validLengths.Contains(this.GetSignificantDigits(isbnCode).Length);
		}

		public bool IsValid(string subject)
		{
			return !this.GetValidationErrors(subject).Any();
		}

		public IEnumerable<string> GetValidationErrors(string subject)
		{
			if (!this.IsLengthValid(subject))
			{
				yield return ValidationMessages.InvalidIsbnLength;
			}

			if (this.StartsWithHyphen(subject))
			{
				yield return ValidationMessages.DontStartIsbnWithHyphens;
			}

			if (!this.HasCorrectHyphenCount(subject))
			{
				yield return this.AreHyphensRequired(subject) ?
					ValidationMessages.InvalidRequiredHyphenCount :
					ValidationMessages.InvalidOptionalHyphenCount;
			}

			if (this.HasConsecutiveHyphens(subject))
			{
				yield return ValidationMessages.DontLetIsbnHaveConsecutiveHyphens;
			}
		}

		private bool IsLengthValid(string subject)
		{
			// We only care about significant digits, so let's strip out the hyphens
			return this.validLengths.Contains(this.GetSignificantDigits(subject).Length);
		}

		private bool StartsWithHyphen(string subject)
		{
			// Could utilise regex for this, but performance is a premium on validation code, and this
			// is simple enough without regexes.
			return subject.First() == '-';
		}

		private bool HasCorrectHyphenCount(string subject)
		{
			int hyphenCount = subject.Where(c => c == '-').Count();

			return this.AreHyphensRequired(subject) ?
				hyphenCount > 0 && this.validHyphenCounts.Contains(hyphenCount) :
				this.validHyphenCounts.Contains(hyphenCount);
		}

		private bool HasConsecutiveHyphens(string subject)
		{
			// Likewise could utilise regex, but there's just no need, and it could impact performance
			return subject.Contains("--");
		}

		private string GetSignificantDigits(string subject)
		{
			return subject.Replace("-", string.Empty);
		}

		private bool AreHyphensRequired(string subject)
		{
			return this.lengthsRequiringHyphens.Contains(this.GetSignificantDigits(subject).Length);
		}
	}
}
